package com.sksamuel.scoot.soa

import org.springframework.beans.factory.annotation.Autowired
import com.googlecode.genericdao.search.{Search => JavaSearch, SearchResult, ISearch}
import org.hibernate.SessionFactory
import com.googlecode.genericdao.dao.hibernate.HibernateBaseDAO
import com.googlecode.genericdao.dao.DAOUtil
import scala.collection.JavaConverters._

/** @author Stephen Samuel */
trait GenericDao[T, ID <: java.io.Serializable] {

    /**
     * <p>
     * Get the entity with the specified type and id from the datastore.
     *
     * <p>
     * If none is found, return null.
     */
    def find(id: ID): T

    /**
     * <p>
     * If the id of the entity is null or zero, add it to the datastore and
     * assign it an id; otherwise, update the corresponding entity in the
     * datastore with the properties of this entity. In either case the entity
     * passed to this method will be attached to the session.
     *
     * <p>
     * If an entity to update is already attached to the session, this method
     * will have no effect. If an entity to update has the same id as another
     * instance already attached to the session, an error will be thrown.
     *
     * @return <code>true</code> if create; <code>false</code> if update.
     */
    def save(entity: T): Boolean

    /**
     * Remove the specified entity from the datastore.
     *
     * @return <code>true</code> if the entity is found in the datastore and
     *         removed, <code>false</code> if it is not found.
     */
    def remove(entity: T): Boolean

    /**
     * Remove the entity with the specified type and id from the datastore.
     *
     * @return <code>true</code> if the entity is found in the datastore and
     *         removed, <code>false</code> if it is not found.
     */
    def removeById(id: ID): Boolean

    /**
     * Get a list of all the objects of the specified type.
     */
    def findAll: List[T]

    /**
     * Get a list of all the objects of the specified type up to the specified limit
     */
    def findAll(limit: Int): List[T]

    /**
     * JavaSearch for entities given the JavaSearch parameters in the specified
     * <code>IJavaSearch</code> obj.
     *
     *
     */
    def search(search: ISearch): List[T]

    /**
     * JavaSearch for a single entity using the given parameters.
     *
     * The result type is automatically determined by the context in which the method is called.
     */
    def searchUnique(search: ISearch): T

    /**
     * Returns the total number of results that would be returned using the
     * given <code>IJavaSearch</code> if there were no paging or maxResults limits.
     */
    def count(search: ISearch): Int

    /**
     * Returns a <code>JavaSearchResult</code> obj that includes both the list of
     * results like <code>JavaSearch()</code> and the total length like
     * <code>count()</code>.
     *
     * The result type is automatically determined by the context in which the method is called.
     */
    def searchAndCount(search: ISearch): SearchResult[T]

}

class GenericDaoImpl[T, ID <: java.io.Serializable] extends HibernateBaseDAO with GenericDao[T, ID] {

    @Autowired
    override def setSessionFactory(sessionFactory: SessionFactory) {
        super.setSessionFactory(sessionFactory)
    }

    protected var persistentClass: Class[T] = DAOUtil
      .getTypeArguments(classOf[GenericDaoImpl[_, _]], this.getClass)
      .get(0)
      .asInstanceOf[Class[T]]

    def count: Int = _count(persistentClass, new JavaSearch)

    def count(JavaSearch: ISearch): Int = {
        _count(persistentClass, JavaSearch)
    }

    def find(id: ID): T = {
        _get(persistentClass, id)
    }

    def findAll: List[T] = {
        _all(persistentClass).asScala.toList
    }

    def findAll(limit: Int): List[T] = {
        search(new JavaSearch(persistentClass).setMaxResults(limit))
    }

    def remove(entity: T): Boolean = {
        _deleteEntity(entity)
    }

    def removeById(id: ID): Boolean = {
        _deleteById(persistentClass, id)
    }

    def save(entity: T): Boolean = {
        _saveOrUpdateIsNew(entity)
    }

    def search(JavaSearch: ISearch): List[T] = {
        _search(persistentClass, JavaSearch).asScala.toList.asInstanceOf[List[T]]
    }

    def searchAndCount(JavaSearch: ISearch): SearchResult[T] = {
        if (JavaSearch == null) {
            val result: SearchResult[T] = new SearchResult[T]
            result.setResult(findAll.asInstanceOf[java.util.List[T]])
            result.setTotalCount(result.getResult.size)
            return result
        }
        _searchAndCount(persistentClass, JavaSearch).asInstanceOf[SearchResult[T]]
    }

    def searchUnique(JavaSearch: ISearch): T = {
        _searchUnique(persistentClass, JavaSearch).asInstanceOf[T]
    }

}

class Search[T: Manifest] extends JavaSearch(manifest.erasure)
