package com.sksamuel.scoot.soa

import reflect.BeanProperty
import org.apache.commons.lang.builder.ToStringBuilder
import javax.servlet.http.HttpServletRequest
import com.fasterxml.jackson.annotation.JsonIgnore
import java.util
import com.fasterxml.jackson.databind.annotation.JsonSerialize
import scala.collection.JavaConverters._

/** @author Stephen Samuel */
@JsonSerialize(include = JsonSerialize.Inclusion.NON_EMPTY)
class PagedQuery {
    @BeanProperty var pageNumber: Int = Page.FirstPage
    @BeanProperty var pageSize: Int = Page.DefaultPageSize
    def offset: Int = (pageNumber - 1) * pageSize
}

object PagedQuery {
    def apply(pageNumber: Int, pageSize: Int) = {
        val q = new PagedQuery
        q.pageNumber = pageNumber
        q.pageSize = pageSize
        q
    }
}

@JsonSerialize(include = JsonSerialize.Inclusion.NON_EMPTY)
class Page[T] {

    @BeanProperty var results: List[T] = _
    @BeanProperty var pageNumber: Int = _
    @BeanProperty var pageSize: Int = _
    @BeanProperty var totalResults: Int = _
    @BeanProperty var numberOfResults: Int = _

    @JsonIgnore def java = if (results == null) new util.ArrayList[T]() else results.asJava
    override def toString = ToStringBuilder.reflectionToString(this)
}

object Page {

    val FirstPage = 1
    val DefaultPageSize = 20

    implicit def java2scala[T](iterable: java.lang.Iterable[T]) = iterable.asScala

    /**
     * Returns an empty Page which pageNumber 1, pageSize 20 and and totalResults 0
     */
    def empty[T]: Page[T] = apply(Nil, FirstPage, DefaultPageSize, 0)

    def apply[T](iterable: java.lang.Iterable[T], pageNumber: Int, pageSize: Int, totalResults: Int): Page[T] =
        apply(iterable.asScala.toList, pageNumber, pageSize, totalResults)

    def apply[T](iterable: java.lang.Iterable[T], req: PagedQuery, totalResults: Int): Page[T] =
        apply(iterable, req.pageNumber, req.pageSize, totalResults)

    def apply[T](array: Array[T]): Page[T] = apply(array, FirstPage)

    def apply[T](array: Array[T], totalResults: Int): Page[T] = apply(array.toList, 1, totalResults, totalResults)

    /**
     * Returns a Page from a given Iterable, as pageNumber 1, where the size of the iterable
     * is taken as the pageSize and the totalResults.
     */
    def apply[T](iterable: Iterable[T]): Page[T] = {
        val list = iterable.toList
        val size = list.size
        apply(list, 1, size, size)
    }

    /**
     * Returns a Page from a given Iterable, as pageNumber 1, where the size of the iterable
     * is taken as the pageSize and the totalResults is as specified.
     */
    def apply[T](iterable: Iterable[T], totalResults: Int): Page[T] =
        apply(iterable, 1, totalResults, totalResults)

    /**
     * Returns a Page from the given Iterable, where the pageSize and pageNumber is taken from the PagedQuery, and the
     * totalResults is as specified.
     */
    def apply[T](iterable: Iterable[T], req: PagedQuery, totalResults: Int): Page[T] =
        apply(iterable, req.pageNumber, req.pageSize, totalResults)

    /**
     * Returns a Page from a given Iterable using the specified values for the pageNumber, pageSize and totalResults.
     */
    def apply[T](iterable: Iterable[T], pageNumber: Int, pageSize: Int, totalResults: Int): Page[T] = {
        val page = new Page[T]
        page.results = iterable.toList
        page.pageNumber = pageNumber
        page.pageSize = pageSize
        page.totalResults = totalResults
        page.numberOfResults = page.results.size
        page
    }
}

class Paging(url: UriBuilder, @BeanProperty val pageNumber: Int, @BeanProperty val totalPages: Int) {
    def before(max: Int): Array[Int] = scala.math.max(1, pageNumber - max).to(pageNumber - 1).toArray
    def after(max: Int): Array[Int] = (pageNumber + 1).to(scala.math.min(pageNumber + max, totalPages)).toArray
    def range(dist: Int): Array[Int] = (before(dist) :+ pageNumber) ++ after(dist)
    def previous: Int = pageNumber - 1
    def next: Int = pageNumber + 1
    def hasPreviousPage: Boolean = pageNumber > 1
    def hasNextPage: Boolean = pageNumber < totalPages
    def isFirstPage: Boolean = pageNumber == 1
    def isLastPage: Boolean = pageNumber == totalPages
    def url(pageNumber: Int): String = (url ? ("pageNumber" -> pageNumber)).toString
    def urlWithOutParams(params: Array[String]): String = {
        var tmp = url
        params.foreach(p => {
            tmp = tmp - (p)
        })
        tmp.toString
    }
    override def toString = ToStringBuilder.reflectionToString(this)
}

object Paging {
    def apply(req: HttpServletRequest, page: Page[_]): Paging =
        apply(req.getRequestURL.toString, Option(req.getQueryString).getOrElse(""), page)
    def apply(url: String, qryString: String, page: Page[_]): Paging = {
        assert(page.pageNumber > 0)
        val totalPages = if (page.pageSize == 0) 0 else scala.math.ceil(page.totalResults.toFloat / page.pageSize).toInt
        var uri = UriBuilder(url)
        if (qryString.trim.length > 0) {
            qryString.split("&").foreach(kv => {
                val arr = kv.split("=")
                if (arr.size == 2) uri = uri &(arr(0), arr(1))
            })
        }
        new Paging(uri, page.pageNumber, totalPages)
    }
}